@@ -1,8 +1,19 @@
 /*
- * Copyright (C) 2002-2003  Pekka Enberg <penberg@iki.fi>
+ * Copyright (C) 2002-2004  Pekka Enberg <penberg@iki.fi>
  *
- * Distributed under the terms of the GNU General Public License
- * version 2 or later.
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 #include <assert.h>
 #include <stdio.h>
@@ -21,6 +32,7 @@ static void end_tag(void);
 static void empty_element_tag(void);
 static void comment(void);
 static void content(void);
+static void processing_insn(void);
 static void do_newline(struct buffer * buffer, const char * text);
 
 #include "lex.yy.c"
@@ -212,18 +224,6 @@ static void force_newline_after_tag(struct buffer * buffer)
     do_unput(current);
 }
 
-/* Force newline for wrapping line. Use lexer to get current character and
-   do not eat whitespace from next line.  */
-static void force_newline_for_wrap(struct buffer * buffer)
-{
-    int current = input();
-
-    if (!is_newline(current)) {
-	do_newline(buffer, "\n");
-    }
-    do_unput(current);
-}
-
 static bool using_primary_buf(void)
 {
     return current_buf == &primary_buf;
@@ -263,6 +263,24 @@ static void merge_buffers(void)
     indent_delta++;
 }
 
+/* Force newline for wrapping line. Use lexer to get current character and
+   do not eat whitespace from next line.  */
+static void force_newline_for_wrap(struct buffer * buffer)
+{
+    int current = input();
+
+    /*
+     * Flush all pending stuff before doing the newline.
+     */
+    if (!using_primary_buf()) {
+	merge_buffers();
+    }
+    do_newline(current_buf, "\n");
+
+    if (!is_newline(current))
+	do_unput(current);
+}
+
 static void start_tag(void)
 {
     char * tmp;
@@ -314,6 +332,7 @@ static void end_tag(void)
 
 static void empty_element_tag(void)
 {
+    bool can_force_newline;
     char * tmp;
     /*
      * Save text because force_newline_before_tag can trash it.
@@ -324,12 +343,19 @@ static void empty_element_tag(void)
        Therefore we use start-tag options before the tag and end-tag
        options after the tag.  */
 
-    if (force_newline_before_start_tag) {
-        force_newline_before_tag(current_buf);
+    if (!using_primary_buf()) {
+	/*
+	 * This is second start-tag. Thus first one has children. We can force
+	 * newline here if we want.
+	 */
+        merge_buffers_start_tag();
+        can_force_newline = force_always;
+    } else {
+        can_force_newline = true;
     }
     buffer_push_str(current_buf, tmp);
 
-    if (force_newline_after_end_tag && force_always) {
+    if (force_newline_after_end_tag && can_force_newline) {
         /* Empty element never has any children. Force newline only if user
            explicitly requested it.  */
         force_newline_after_tag(current_buf);
@@ -400,7 +426,11 @@ static unsigned long indent_size(void)
 
 static bool need_wrap(struct buffer * buffer)
 {
-    return buffer_size(buffer) + indent_size() == max_columns;
+    if (buffer == &primary_buf)
+	return buffer_size(buffer) + indent_size() == max_columns;
+    else
+	return (buffer_size(&primary_buf) + buffer_size(&tag_buf)
+		+ buffer_size(buffer) + indent_size()) >= max_columns;
 }
 
 static void content(void)
@@ -460,3 +490,8 @@ static void content(void)
 	buffer_release(&tmp);
     }
 }
+
+static void processing_insn(void)
+{
+    buffer_push_str(current_buf, yytext);
+}
